/*
Copyright (c) 2010, Intel Corporation
All rights reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    * Redistributions of source code must retain the above copyright notice,
      this list of conditions and the following disclaimer.
    * Redistributions in binary form must reproduce the above copyright notice,
      this list of conditions and the following disclaimer in the documentation
      and/or other materials provided with the distribution.
    * Neither the name of Intel Corporation nor the names of its contributors
      may be used to endorse or promote products derived from this software
      without specific prior written permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR
ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
(INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON
ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS
SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*/

#include <grub/extcmd.h>
#include <grub/types.h>
#include <grub/misc.h>
#include <grub/mm.h>
#include <grub/env.h>
#include <grub/err.h>
#include <grub/dl.h>
#include <grub/file.h>
#include <grub/term.h>
#include <grub/time.h>
#include "c.h"
#include "datatype.h"
#include "bitsutil.h"
#include "pci.h"
#include "smp.h"

enum verbose_level {
    V_NONE,
    V_FAIL,
    V_DETAIL,
    V_PASS,
};

static enum verbose_level verbose = V_FAIL;
static U32 ncpus = 0;
static const CPU_INFO *cpu;
#define ALL_CPUS (~0UL)
static U32 viewpoint = ALL_CPUS;
static unsigned passed = 0, failed = 0;

struct dword_regs {
    U32 eax;
    U32 ebx;
    U32 ecx;
    U32 edx;
};

struct unique_cpuid {
    U32 count;
    struct dword_regs regs;
};

static struct dword_regs *cpuid = 0;
static struct unique_cpuid *unique_cpuid = 0;

struct mem {
    U32 addr;
    U32 value;
};

struct unique_mem {
    U32 count;
    struct mem mem;
};

static struct mem *mem = 0;
static struct unique_mem *unique_mem = 0;

struct msr {
    U32 num;
    U32 status;
    U64 value;
};

struct unique_msr {
    U32 count;
    struct msr msr;
};

static struct msr *msr = 0;
static struct unique_msr *unique_msr = 0;

struct brand {
    char string[48];
};

struct unique_brand {
    U32 count;
    char string[48];
};

static struct brand *brand = 0;
static struct unique_brand *unique_brand = 0;

static U32 *cpu_status = 0;

static U32 unique_count;

static grub_err_t parse_cpu_num(char *str, U32 * out)
{
    U32 num;

    grub_errno = GRUB_ERR_NONE;

    if (grub_strcmp(str, "all") == 0) {
        *out = ALL_CPUS;
        return GRUB_ERR_NONE;
    }

    if (strtou32_h(str, &num) != GRUB_ERR_NONE)
        return grub_errno;

    if (num >= ncpus)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "CPU number %u invalid", num);

    *out = num;
    return GRUB_ERR_NONE;
}

static U32 first_cpu(U32 cpu_num)
{
    return cpu_num == ALL_CPUS ? 0 : cpu_num;
}

static U32 last_cpu(U32 cpu_num)
{
    return cpu_num == ALL_CPUS ? ncpus : cpu_num + 1;
}

static void msr_callback(void *param)
{
    struct msr *m = param;
    rdmsr64(m->num, &m->value, &m->status);
}

static bool msr_match(struct msr msr1, struct msr msr2)
{
    return ((msr1.status == msr2.status) && (msr1.status || msr1.value == msr2.value));
}

static void get_msr(U32 cpu_num, U32 num, U64 mask, U32 shift)
{
    U32 i, j;

    unique_count = 0;

    for (i = first_cpu(cpu_num); i != last_cpu(cpu_num); i++) {
        bool found = false;
        msr[i].num = num;
        msr[i].status = 0xdeadbeef;
        if (!smp_function(cpu[i].apicid, msr_callback, &msr[i]))
            grub_printf("DEBUG: smp_function failed in get_msr\n");
        msr[i].value >>= shift;
        msr[i].value &= mask;

        for (j = 0; j < unique_count; j++)
            if (msr_match(unique_msr[j].msr, msr[i])) {
                unique_msr[j].count++;
                found = true;
                break;
            }
        if (!found) {
            unique_msr[unique_count].count = 1;
            unique_msr[unique_count].msr = msr[i];
            unique_count++;
        }
    }
}

static bool msr_mask_value(const char *teststr, U32 num, U64 mask, U64 value, U32 shift)
{
    U32 i, j;
    bool status = true;
    char msr_desc[128];

    get_msr(ALL_CPUS, num, mask, shift);

    for (i = 0; i < unique_count; i++) {
        if (mask == ~0ULL) {
            if (shift == 0)
                grub_snprintf(msr_desc, sizeof(msr_desc), "MSR 0x%x == 0x%llx", num, value);
            else
                grub_snprintf(msr_desc, sizeof(msr_desc), "MSR 0x%x shr %u == 0x%llx", num, shift, value);
        } else {
            if (shift == 0)
                grub_snprintf(msr_desc, sizeof(msr_desc), "MSR 0x%x & 0x%llx == 0x%llx", num, mask, value);
            else
                grub_snprintf(msr_desc, sizeof(msr_desc), "MSR 0x%x shr %u & 0x%llx == 0x%llx", num, shift, mask, value);
        }

        if (unique_msr[i].msr.status) {
            if (verbose >= V_DETAIL)
                grub_printf("%s (%s): FAIL (GPF)\n", teststr, msr_desc);
            else if (verbose)
                grub_printf("%s FAIL\n", teststr);
            status = false;
        } else if (unique_msr[i].msr.value != value) {
            if (verbose >= V_DETAIL)
                grub_printf("%s (%s): FAIL (0x%llx)\n", teststr, msr_desc, unique_msr[i].msr.value);
            else if (verbose)
                grub_printf("%s FAIL\n", teststr);
            status = false;
        } else if (verbose >= V_PASS)
            grub_printf("%s (%s): PASS\n", teststr, msr_desc);
        else
            continue;

        if (verbose >= V_DETAIL) {
            grub_printf("On %u CPUs:%s", unique_msr[i].count, unique_msr[i].count == ncpus ? " all" : "");
            if (unique_msr[i].count != ncpus)
                for (j = 0; j < ncpus; j++)
                    if (msr_match(unique_msr[i].msr, msr[j]))
                        grub_printf(" 0x%x", cpu[j].apicid);
            grub_printf("\n");
        }
    }

    if (status)
        passed++;
    else
        failed++;

    return status;
}

static bool msr_consistent(const char *teststr, U32 num, U64 mask, U32 shift)
{
    U32 i, j;
    bool status;

    get_msr(ALL_CPUS, num, mask, shift);

    if (unique_count == 1) {
        passed++;
        status = true;
    } else {
        failed++;
        status = false;
    }

    if ((verbose && !status) || (verbose >= V_PASS)) {
        if (*teststr)
            grub_printf("%s ", teststr);

        if (verbose >= V_DETAIL || !*teststr) {
            grub_printf("(MSR 0x%x ", num);
            if (mask != ~0ULL)
                grub_printf("& 0x%llx", mask);
            grub_printf("consistent): ");
        }
        grub_printf(" %s", status ? "PASS" : "FAIL");
        if (verbose >= V_DETAIL)
            grub_printf(" (%u different values)", unique_count);
        grub_printf("\n");

        if (verbose >= V_DETAIL) {
            for (i = 0; i < unique_count; i++) {
                grub_printf("%u CPUs ", unique_msr[i].count);
                if (unique_msr[i].msr.status != 0 && unique_msr[i].msr.status != ~0U)
                    grub_printf("DEBUG: unique_msr[%u] not 0 or ~0: 0x%x\n", i, unique_msr[i].msr.status);
                if (unique_msr[i].msr.status)
                    grub_printf("GPF:               ");
                else
                    grub_printf("0x%016llx: ", unique_msr[i].msr.value);

                for (j = 0; j < ncpus; j++)
                    if (msr_match(unique_msr[i].msr, msr[j]))
                        grub_printf(" 0x%x", cpu[j].apicid);
                grub_printf("\n");
            }
        }
    }

    return status;
}

static bool pci_mask_value(const char *teststr, U8 bus, U8 dev, U8 fn, U8 reg, U8 size, U32 mask, U32 shift, U32 expected_value, U32 * regval)
{
    char desc[128];
    U32 value;
    bool status;

    if (mask == ~0U) {
        if (shift == 0)
            grub_snprintf(desc, sizeof(desc), "PCI %02x:%02x.%01x reg 0x%02x == 0x%x", bus, dev, fn, reg, expected_value);
        else
            grub_snprintf(desc, sizeof(desc), "PCI %02x:%02x.%01x reg 0x%02x shr %u == 0x%x", bus, dev, fn, reg, shift, expected_value);
    } else {
        if (shift == 0)
            grub_snprintf(desc, sizeof(desc), "PCI %02x:%02x.%01x reg 0x%02x & 0x%x == 0x%x", bus, dev, fn, reg, mask, expected_value);
        else
            grub_snprintf(desc, sizeof(desc), "PCI %02x:%02x.%01x reg 0x%02x shr %u & 0x%x == 0x%x", bus, dev, fn, reg, shift, mask, expected_value);
    }

    switch (size) {
    case 1:
        value = pci_read_u8(bus, dev, fn, reg);
        break;
    case 2:
        value = pci_read_u16(bus, dev, fn, reg);
        break;
    case 4:
        value = pci_read_u32(bus, dev, fn, reg);
        break;
    default:
        return false; /* Don't do that. */
    }

    *regval = value = (value >> shift) & mask;
    status = (value == expected_value);
    if (status)
        passed++;
    else
        failed++;
    if ((verbose && !status) || (verbose >= V_PASS)) {
        if (*teststr)
            grub_printf("%s ", teststr);
        if (verbose >= V_DETAIL || !*teststr) {
            if (!status)
                grub_printf("(%s) (%x): ", desc, value);
            else
                grub_printf("(%s): ", desc);
        }
        grub_printf("%s\n", status ? "PASS" : "FAIL");
    }
    return status;
}

static bool pcie_mask_value(const char *teststr, U32 memaddr, U8 bus, U8 dev, U8 fn, U8 reg, U8 size, U32 mask, U32 shift, U32 expected_value, U32 * regval)
{
    char desc[128];
    U32 value;
    bool status;

    if (mask == ~0U) {
        if (shift == 0)
            grub_snprintf(desc, sizeof(desc), "PCIe %02x:%02x.%01x reg 0x%02x == 0x%x", bus, dev, fn, reg, expected_value);
        else
            grub_snprintf(desc, sizeof(desc), "PCIe %02x:%02x.%01x reg 0x%02x shr %u == 0x%x", bus, dev, fn, reg, shift, expected_value);
    } else {
        if (shift == 0)
            grub_snprintf(desc, sizeof(desc), "PCIe %02x:%02x.%01x reg 0x%02x & 0x%x == 0x%x", bus, dev, fn, reg, mask, expected_value);
        else
            grub_snprintf(desc, sizeof(desc), "PCIe %02x:%02x.%01x reg 0x%02x shr %u & 0x%x == 0x%x", bus, dev, fn, reg, shift, mask, expected_value);
    }

    switch (size) {
    case 1:
        value = pci_mmcfg_read_u8(memaddr, bus, dev, fn, reg);
        break;
    case 2:
        value = pci_mmcfg_read_u16(memaddr, bus, dev, fn, reg);
        break;
    case 4:
        value = pci_mmcfg_read_u32(memaddr, bus, dev, fn, reg);
        break;
    default:
        return false;
    }

    *regval = value = (value >> shift) & mask;
    status = (value == expected_value);
    if (status)
        passed++;
    else
        failed++;
    if ((verbose && !status) || (verbose >= V_PASS)) {
        if (*teststr)
            grub_printf("%s ", teststr);
        if (verbose >= V_DETAIL || !*teststr) {
            if (!status)
                grub_printf("(%s) (%x): ", desc, value);
            else
                grub_printf("(%s): ", desc);
        }
        grub_printf("%s\n", status ? "PASS" : "FAIL");
    }
    return status;
}

static void uninit(void)
{
    grub_free(cpuid);
    grub_free(unique_cpuid);
    grub_free(msr);
    grub_free(unique_msr);
    grub_free(brand);
    grub_free(unique_brand);
    grub_free(cpu_status);
}

static grub_err_t oom_uninit(void)
{
    uninit();
    return grub_error(GRUB_ERR_OUT_OF_MEMORY, "Out of memory");
}

static grub_err_t init(void)
{
    uninit();

    ncpus = smp_init();
    if (!ncpus)
        return grub_error(GRUB_ERR_IO, "Failed to initialize SMP");

    cpu = smp_read_cpu_list();
    if (!cpu)
        return grub_error(GRUB_ERR_IO, "Failed to initialize SMP (smp_read_cpu_list)");

    cpuid = grub_malloc(ncpus * sizeof(*cpuid));
    if (!cpuid)
        return oom_uninit();

    unique_cpuid = grub_malloc(ncpus * sizeof(*unique_cpuid));
    if (!unique_cpuid)
        return oom_uninit();

    msr = grub_malloc(ncpus * sizeof(*msr));
    if (!msr)
        return oom_uninit();

    unique_msr = grub_malloc(ncpus * sizeof(*unique_msr));
    if (!unique_msr)
        return oom_uninit();

    brand = grub_malloc(ncpus * sizeof(*brand));
    if (!brand)
        return oom_uninit();

    unique_brand = grub_malloc(ncpus * sizeof(*unique_brand));
    if (!unique_brand)
        return oom_uninit();

    cpu_status = grub_malloc(ncpus * sizeof(*cpu_status));
    if (!cpu_status)
        return oom_uninit();

    mem = grub_malloc(ncpus * sizeof(*mem));
    if (!mem)
        return oom_uninit();

    unique_mem = grub_malloc(ncpus * sizeof(*unique_mem));
    if (!unique_mem)
        return oom_uninit();

    return GRUB_ERR_NONE;
}

static void cpuid_callback(void *param)
{
    struct dword_regs *r = param;
    cpuid32_indexed(r->eax, r->ecx, &r->eax, &r->ebx, &r->ecx, &r->edx);
}

static void brandstring_callback(void *param)
{
    U32 i;
    U32 nleaves, dummy;
    U32 *dwords = param;

    cpuid32(0x80000000, &nleaves, &dummy, &dummy, &dummy);
    if (nleaves < 0x80000004) {
        dwords[0] = 0;
        return;
    }

    for (i = 0x80000002; i <= 0x80000004; i++, dwords += 4)
        cpuid32(i, dwords, dwords + 1, dwords + 2, dwords + 3);
}

static void get_brandstring(U32 cpu_num)
{
    U32 i, j;

    unique_count = 0;

    for (i = first_cpu(cpu_num); i != last_cpu(cpu_num); i++) {
        bool found = false;

        if (!smp_function(cpu[i].apicid, brandstring_callback, brand[i].string))
            grub_printf("DEBUG: smp_function failed in get_brandstring\n");

        for (j = 0; j < unique_count; j++)
            if (!grub_memcmp(unique_brand[j].string, brand[i].string, sizeof(*brand))) {
                unique_brand[j].count++;
                found = true;
                break;
            }
        if (!found) {
            unique_brand[unique_count].count = 1;
            grub_memcpy(unique_brand[unique_count].string, brand[i].string, sizeof(*brand));
            unique_count++;
        }
    }
}

static void brandstring_values(U32 cpu_num)
{
    U32 i, j;

    get_brandstring(cpu_num);

    for (i = 0; i < unique_count; i++) {
        grub_printf("brandstring: \"%s\"\n", unique_brand[i].string);
        grub_printf("On %u CPUs:%s", unique_brand[i].count, unique_brand[i].count == ncpus ? " all" : "");
        if (unique_brand[i].count != ncpus)
            for (j = first_cpu(cpu_num); j < last_cpu(cpu_num); j++)
                if (!grub_memcmp(unique_brand[i].string, brand[j].string, sizeof(*brand)))
                    grub_printf(" 0x%x", cpu[j].apicid);
        grub_printf("\n");
    }
}

static const struct grub_arg_option brandstring_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_brandstring(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 cpuNum = viewpoint;
    (void)argc;
    (void)args;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    brandstring_values(cpuNum);

    return GRUB_ERR_NONE;
}

static const char c_help[] =
    "Usage: c \"C-style expression with space-separated tokens\"\n"
    "Evaluate a C expression\n"
    "All evaluation occurs on 64-bit unsigned integers, specified as decimal\n"
    "numbers, hex numbers prefixed by 0x, or named variables taken from the\n"
    "environment.  Variable assignments work, and set variables in the environment.\n"
    "Returns true or false based on the final value of the expression, so\n"
    "\"if c ... \" and \"while c ...\" work.\n"
    "\n"
    "Supported operators, in order from highest to lowest precedence:\n"
    " ()        (parentheses)\n"
    " ! ~ ++ -- (both pre- and post- increment and decrement)\n"
    " * / %\n"
    " + -\n"
    " << >>\n"
    " < <= > >=\n"
    " == !=\n"
    " &         (bitwise and)\n"
    " ^         (bitwise xor)\n"
    " |         (bitwise or)\n"
    " &&        (logical and; WARNING: does not short-circuit)\n"
    " ||        (logical or;  WARNING: does not short-circuit)\n"
    " = += -= *= /= %= <<= >>= &= ^= |=\n"
    " ,         (comma operator)\n";

static grub_err_t grub_cmd_c(grub_command_t cmd, int argc, char **args)
{
    U64 result;
    (void)cmd;

    if (argc == 1 && grub_strcmp(args[0], "--help") == 0) {
        grub_printf("%s", c_help);
        return GRUB_ERR_NONE;
    }

    if (c_expr(argc, args, &result))
        return !result;
    else
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Failed to parse C expression");
}

static const struct grub_arg_option cpu_options[] = {
#undef OPTION_LIST
#define OPTION_LIST 0
    {"list", 'l', 0, "List CPU numbers", 0, 0},
#undef OPTION_ENV
#define OPTION_ENV 1
    {"env", 'e', 0, "Set environment vars cpu_list and apicid_list (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_cpu(struct grub_extcmd *cmd, int argc, char **args)
{
    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (argc > 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 0 or 1 argument: [cpu_number]");
    else if (argc == 1)
        if (parse_cpu_num(args[0], &viewpoint) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_LIST].set) {
        U32 i;
        grub_printf("%u CPUs: 0-%u\n", ncpus, ncpus - 1);
        for (i = 0; i < ncpus; i++)
            grub_printf("CPU%3u = 0x%02x%s", i, cpu[i].apicid, (i != ncpus - 1) && (i + 1) % 4 ? "\t" : "\n");
    }
    if (viewpoint == ALL_CPUS)
        grub_printf("Assigned CPU number = all\n");
    else
        grub_printf("Assigned CPU number = %u\n", viewpoint);

    if (cmd->state[OPTION_ENV].set) {
        U32 i;
        char *apicid_list;
        char *cpu_list;

        apicid_list = (char *)grub_malloc(ncpus * sizeof("0xXXXXXXXX "));
        /* FIXME: update if we remove MAX_LOGICAL_CPU and make it dynamic */
        cpu_list = (char *)grub_malloc(ncpus * sizeof("XXX "));
        if (!apicid_list || !cpu_list) {
            grub_free(apicid_list);
            grub_free(cpu_list);
            return grub_error(GRUB_ERR_OUT_OF_MEMORY, "Out of memory");
        }
        *apicid_list = *cpu_list = 0;
        for (i = 0; i < ncpus; i++) {
            char *cpu_str = grub_xasprintf("%s%u", i == 0 ? "" : " ", i);
            char *apicid_str = grub_xasprintf("%s0x%x", i == 0 ? "" : " ", cpu[i].apicid);
            grub_strcat(cpu_list, cpu_str);
            grub_strcat(apicid_list, apicid_str);
            grub_free(cpu_str);
            grub_free(apicid_str);
        }
        grub_env_set("cpu_list", cpu_list);
        grub_env_set("apicid_list", apicid_list);
        grub_free(cpu_list);
        grub_free(apicid_list);
    }
    return GRUB_ERR_NONE;
}

static void print_hz(U64 hz)
{
    U64 temp64;
    U32 temp32;

    temp64 = grub_divmod64(hz, 1000 * 1000 * 1000, &temp32);
    if (temp64 > 0)
        grub_printf("%llu.%03u GHz", temp64, temp32 / (1000 * 1000));
    else {
        temp64 = grub_divmod64(hz, 1000 * 1000, &temp32);
        if (temp64 > 0)
            grub_printf("%llu.%03u MHz", temp64, temp32 / 1000);
        else
            grub_printf("%llu Hz", hz);
    }
}

#define IA32_MPERF_MSR 0xE7
#define IA32_APERF_MSR 0xE8

static grub_err_t grub_cmd_cpu_freq(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 eax, ecx, dummy;
    U64 mperf, aperf, delta_hz;
    U32 status;
    U64 start, stop;
    U64 mperf_hz, aperf_hz;

    (void)cmd;
    (void)argc;
    (void)args;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    cpuid32(0, &eax, &dummy, &dummy, &dummy);
    if (eax < 6) {
        grub_printf("CPUID Leaf 6 is not supported\n");
        return GRUB_ERR_NONE;
    }

    cpuid32(6, &dummy, &dummy, &ecx, &dummy);
    if (!(ecx & (1 << 0))) {
        grub_printf("MPERF/APERF MSRs are not supported\n");
        return GRUB_ERR_NONE;
    }

    wrmsr64(IA32_MPERF_MSR, 0, &status);
    if (status) {
        grub_printf("Writing of IA32_MPERF MSR caused a GPF\n");
        return GRUB_ERR_NONE;
    }
    wrmsr64(IA32_APERF_MSR, 0, &status);
    if (status) {
        grub_printf("Writing of IA32_APERF MSR caused a GPF\n");
        return GRUB_ERR_NONE;
    }
    start = grub_get_time_ms();

    while (grub_get_time_ms() - start < 1000) {
    }

    rdmsr64(IA32_MPERF_MSR, &mperf, &status);
    if (status) {
        grub_printf("Reading of IA32_MPERF MSR caused a GPF\n");
        return GRUB_ERR_NONE;
    }
    rdmsr64(IA32_APERF_MSR, &aperf, &status);
    if (status) {
        grub_printf("Reading of IA32_APERF MSR caused a GPF\n");
        return GRUB_ERR_NONE;
    }
    stop = grub_get_time_ms();

    mperf_hz = grub_divmod64(mperf, stop - start, 0) * 1000;
    aperf_hz = grub_divmod64(aperf, stop - start, 0) * 1000;
    delta_hz = aperf_hz > mperf_hz ? aperf_hz - mperf_hz : mperf_hz - aperf_hz;

    grub_printf("Frequency = ");
    print_hz(mperf_hz);
    grub_printf(" (MPERF)   ");

    print_hz(aperf_hz);
    grub_printf(" (APERF)   ");

    grub_printf("%s", aperf_hz > mperf_hz ? "+" : mperf_hz - aperf_hz ? "-" : "");
    print_hz(delta_hz);
    grub_printf(" (delta)\n");

    return GRUB_ERR_NONE;
}

static bool cpuid_match(struct dword_regs r1, struct dword_regs r2)
{
    return r1.eax == r2.eax && r1.ebx == r2.ebx && r1.ecx == r2.ecx && r1.edx == r2.edx;
}

static void get_cpuid(U32 cpu_num, U32 num, U32 idx, U32 mask_eax, U32 mask_ebx, U32 mask_ecx, U32 mask_edx, U32 shift)
{
    U32 i, j;

    unique_count = 0;

    for (i = first_cpu(cpu_num); i != last_cpu(cpu_num); i++) {
        bool found = false;
        cpuid[i].eax = num;
        cpuid[i].ecx = idx;
        if (!smp_function(cpu[i].apicid, cpuid_callback, &cpuid[i]))
            grub_printf("DEBUG: smp_function failed in get_cpuid\n");

        cpuid[i].eax = (cpuid[i].eax >> shift) & mask_eax;
        cpuid[i].ebx = (cpuid[i].ebx >> shift) & mask_ebx;
        cpuid[i].ecx = (cpuid[i].ecx >> shift) & mask_ecx;
        cpuid[i].edx = (cpuid[i].edx >> shift) & mask_edx;

        for (j = 0; j < unique_count; j++)
            if (cpuid_match(unique_cpuid[j].regs, cpuid[i])) {
                unique_cpuid[j].count++;
                found = true;
                break;
            }
        if (!found) {
            unique_cpuid[unique_count].count = 1;
            unique_cpuid[unique_count].regs = cpuid[i];
            unique_count++;
        }
    }
}

static const struct grub_arg_option cpuid32_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_ENV
#define OPTION_ENV 1
    {"env", 'e', 0, "Set environment vars eax,ebx,ecx,edx (default=disabled)", 0, 0},
#undef OPTION_QUIET
#define OPTION_QUIET 2
    {"quiet", 'q', 0, "No display of register values (default=disabled)", 0, 0},
#undef OPTION_MASK
#define OPTION_MASK 3
    {"mask", 'm', 0, "Default mask to apply to values read (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_MASK_EAX
#define OPTION_MASK_EAX 4
    {"eax-mask", 'A', 0, "Mask to apply to EAX; overrides --mask (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_MASK_EBX
#define OPTION_MASK_EBX 5
    {"ebx-mask", 'B', 0, "Mask to apply to EBX; overrides --mask (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_MASK_ECX
#define OPTION_MASK_ECX 6
    {"ecx-mask", 'C', 0, "Mask to apply to ECX; overrides --mask (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_MASK_EDX
#define OPTION_MASK_EDX 7
    {"edx-mask", 'D', 0, "Mask to apply to EDX; overrides --mask (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 8
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0},
};

static grub_err_t grub_cmd_cpuid32(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 num, idx = 0;
    U32 cpuNum = viewpoint;
    U32 mask = ~0;
    U32 mask_eax, mask_ebx, mask_ecx, mask_edx;
    U32 shift = 0;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    mask_eax = mask_ebx = mask_ecx = mask_edx = mask;

    if (cmd->state[OPTION_MASK_EAX].set)
        if (strtou32_h(cmd->state[OPTION_MASK_EAX].arg, &mask_eax) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK_EBX].set)
        if (strtou32_h(cmd->state[OPTION_MASK_EBX].arg, &mask_ebx) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK_ECX].set)
        if (strtou32_h(cmd->state[OPTION_MASK_ECX].arg, &mask_ecx) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK_EDX].set)
        if (strtou32_h(cmd->state[OPTION_MASK_EDX].arg, &mask_edx) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc < 1 || argc > 2)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 or 2 arguments: function [index]");
    if (strtou32_h(args[0], &num) != GRUB_ERR_NONE)
        return grub_errno;

    if (argc == 2 && strtou32_h(args[1], &idx) != GRUB_ERR_NONE)
        return grub_errno;

    get_cpuid(cpuNum, num, idx, mask_eax, mask_ebx, mask_ecx, mask_edx, shift);

    if (!cmd->state[OPTION_QUIET].set) {
        U32 i, j;
        grub_printf("CPUID function 0x%x", num);
        if (argc == 2)
            grub_printf(" index 0x%x", idx);
        grub_printf("\n");

        if (mask_eax != ~0ul || mask_ebx != ~0ul || mask_ecx != ~0ul || mask_edx != ~0ul || shift != 0)
            grub_printf("Register values have been shifted/masked as specified on the command line\n");

        for (i = 0; i < unique_count; i++) {
            struct dword_regs *r = &unique_cpuid[i].regs;
            grub_printf("eax=%08x  ebx=%08x  ecx=%08x  edx=%08x\n", r->eax, r->ebx, r->ecx, r->edx);
            grub_printf("On %u CPUs:%s", unique_cpuid[i].count, unique_cpuid[i].count == ncpus ? " all" : "");
            if (unique_cpuid[i].count != ncpus)
                for (j = first_cpu(cpuNum); j < last_cpu(cpuNum); j++)
                    if (cpuid_match(unique_cpuid[i].regs, cpuid[j]))
                        grub_printf(" 0x%x", cpu[j].apicid);
            grub_printf("\n");
        }
    }

    if (cmd->state[OPTION_ENV].set) {
        char buf[sizeof("0xXXXXXXXX")];
        struct dword_regs *r = &unique_cpuid[0].regs;
        if (unique_count != 1)
            if (!cmd->state[OPTION_QUIET].set)
                return grub_error(GRUB_ERR_IO, "Variable setting requested but CPUID values not unique");
        grub_snprintf(buf, sizeof(buf), "0x%x", r->eax);
        grub_env_set("eax", buf);
        grub_snprintf(buf, sizeof(buf), "0x%x", r->ebx);
        grub_env_set("ebx", buf);
        grub_snprintf(buf, sizeof(buf), "0x%x", r->ecx);
        grub_env_set("ecx", buf);
        grub_snprintf(buf, sizeof(buf), "0x%x", r->edx);
        grub_env_set("edx", buf);
    }

    return GRUB_ERR_NONE;
}

static const struct grub_arg_option test_cpuid_consistent_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_MASK
#define OPTION_MASK 1
    {"mask", 'm', 0, "Default mask to apply to values read (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_MASK_EAX
#define OPTION_MASK_EAX 2
    {"eax-mask", 'A', 0, "Mask to apply to EAX; overrides --mask (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_MASK_EBX
#define OPTION_MASK_EBX 3
    {"ebx-mask", 'B', 0, "Mask to apply to EBX; overrides --mask (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_MASK_ECX
#define OPTION_MASK_ECX 4
    {"ecx-mask", 'C', 0, "Mask to apply to ECX; overrides --mask (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_MASK_EDX
#define OPTION_MASK_EDX 5
    {"edx-mask", 'D', 0, "Mask to apply to EDX; overrides --mask (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 6
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0},
};

static grub_err_t grub_cmd_test_cpuid_consistent(struct grub_extcmd *cmd, int argc, char **args)
{
    char *teststr;
    U32 num, idx = 0;
    U32 cpuNum = viewpoint;
    U32 mask = ~0;
    U32 mask_eax, mask_ebx, mask_ecx, mask_edx;
    U32 shift = 0;
    bool status;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    mask_eax = mask_ebx = mask_ecx = mask_edx = mask;

    if (cmd->state[OPTION_MASK_EAX].set)
        if (strtou32_h(cmd->state[OPTION_MASK_EAX].arg, &mask_eax) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK_EBX].set)
        if (strtou32_h(cmd->state[OPTION_MASK_EBX].arg, &mask_ebx) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK_ECX].set)
        if (strtou32_h(cmd->state[OPTION_MASK_ECX].arg, &mask_ecx) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK_EDX].set)
        if (strtou32_h(cmd->state[OPTION_MASK_EDX].arg, &mask_edx) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc < 2 || argc > 3)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 2 or 3 arguments: \"description\" function [index]");
    teststr = args[0];
    if (strtou32_h(args[1], &num) != GRUB_ERR_NONE)
        return grub_errno;

    if (argc == 3 && strtou32_h(args[2], &idx) != GRUB_ERR_NONE)
        return grub_errno;

    get_cpuid(cpuNum, num, idx, mask_eax, mask_ebx, mask_ecx, mask_edx, shift);

    if (unique_count == 1) {
        status = true;
        passed++;
    } else {
        status = false;
        failed++;
    }

    if ((verbose && !status) || (verbose >= V_PASS)) {
        U32 i, j;
        if (grub_strlen(teststr))
            grub_printf("%s", teststr);
        else {
            grub_printf("CPUID function 0x%x", num);
            if (argc == 2)
                grub_printf(" index 0x%x", idx);
            grub_printf(" consistent check");
        }
        grub_printf(" %s\n", status ? "PASS" : "FAIL");

        if (verbose >= V_DETAIL) {
            if (mask_eax != ~0ul || mask_ebx != ~0ul || mask_ecx != ~0ul || mask_edx != ~0ul || shift != 0)
                grub_printf("Register values have been shifted/masked as specified on the command line\n");
            for (i = 0; i < unique_count; i++) {
                struct dword_regs *r = &unique_cpuid[i].regs;
                grub_printf("eax=%08x  ebx=%08x  ecx=%08x  edx=%08x\n", r->eax, r->ebx, r->ecx, r->edx);
                grub_printf("On %u CPUs:%s", unique_cpuid[i].count, unique_cpuid[i].count == ncpus ? " all" : "");
                if (unique_cpuid[i].count != ncpus)
                    for (j = first_cpu(cpuNum); j < last_cpu(cpuNum); j++)
                        if (cpuid_match(unique_cpuid[i].regs, cpuid[j]))
                            grub_printf(" 0x%x", cpu[j].apicid);
                grub_printf("\n");
            }
        }
    }

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

#define LATENCY_RECENT_COUNT 6

struct latency_bin {
    const char *desc;
    U64 max;
    U64 avg;
    U64 count;
    U32 recent_index;
    U64 recent_delta[LATENCY_RECENT_COUNT];
    U64 recent_absolute[LATENCY_RECENT_COUNT];
};

#define IA32_TSC_MSR 0x10
/* rdtsc() assumes TSC exists; detect that first by calling rdmsr64 yourself
 * and checking status. */
static U64 rdtsc(void)
{
    U64 temp;
    U32 status;
    rdmsr64(IA32_TSC_MSR, &temp, &status);
    return temp;
}

#define MSR_SMI_COUNT 0x34

static void test_latency(U64 test_duration_tscs, U64 * avg_ptr, U64 * count_ptr, U64 * max_ptr, struct latency_bin *bin, U32 bin_len)
{
    U64 test_start;
    U64 tsc1, tsc2;
    U64 smi_count1, smi_count2;
    U64 current;
    U32 i, status;
    U64 count = 0, sum = 0, max = 0;
    U32 smi_count_msr_valid = 0;

    msr[0].num = MSR_SMI_COUNT;
    if (smp_function(cpu[0].apicid, msr_callback, &msr[0]))
        smi_count_msr_valid = !msr[0].status;

    if (smi_count_msr_valid)
        rdmsr64(MSR_SMI_COUNT, &smi_count1, &status);

    for (test_start = tsc1 = tsc2 = rdtsc(); tsc2 - test_start < test_duration_tscs; tsc1 = tsc2, tsc2 = rdtsc()) {
        current = tsc2 - tsc1;

        for (i = 0; i < bin_len; i++)
            if (current <= bin[i].max) {
                bin[i].count++;
                bin[i].avg += current;
                if (bin[i].recent_index != LATENCY_RECENT_COUNT) {
                    bin[i].recent_delta[bin[i].recent_index] = current;
                    bin[i].recent_absolute[bin[i].recent_index] = tsc2;
                    bin[i].recent_index++;
                }
                break;
            }

        if (current > max)
            max = current;
    }

    if (smi_count_msr_valid) {
        rdmsr64(MSR_SMI_COUNT, &smi_count2, &status);
        // For now, assuming that if this MSR exists then it represents the SMI count
        if (!status)
            grub_printf("%llu SMI detected using MSR_SMI_COUNT (MSR 0x%x)\n", smi_count2 - smi_count1, MSR_SMI_COUNT);
    }

    for (i = 0; i < bin_len; i++) {
        sum += bin[i].avg;
        count += bin[i].count;
        if (bin[i].count)
            bin[i].avg = grub_divmod64(bin[i].avg, bin[i].count, 0);
    }
    *avg_ptr = grub_divmod64(sum, count, 0);
    *count_ptr = count;
    *max_ptr = max;
}

static void print_time(U64 tsc, U64 tsc_per_msec)
{
    U64 temp;
    if ((temp = grub_divmod64(tsc * 1000 * 1000, tsc_per_msec, 0)) < 10 * 1000)
        grub_printf("%4lluns", temp);
    else if ((temp = grub_divmod64(tsc * 1000, tsc_per_msec, 0)) < 10 * 1000)
        grub_printf("%4lluus", temp);
    else if ((temp = grub_divmod64(tsc, tsc_per_msec, 0)) < 10 * 1000)
        grub_printf("%4llums", temp);
    else
        grub_printf("%4llus", grub_divmod64(tsc, 1000 * tsc_per_msec, 0));
}

static grub_err_t grub_cmd_cpu_latency(struct grub_extcmd *cmd, int argc, char **args)
{
#define LATENCY_BINS 10
    struct latency_bin bin[LATENCY_BINS];

    U32 i, j;
    U64 avg, count, max;
    U32 status;
    U64 tsc1, tsc2, start, stop, tsc_per_msec;
    U64 bin_time;

    (void)cmd;
    (void)argc;
    (void)args;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    rdmsr64(IA32_TSC_MSR, &tsc1, &status);
    if (status)
        return grub_error(GRUB_ERR_IO, "Reading of IA32_TSC MSR caused a GPF");

    grub_printf("Starting pass #1.  Calibrating the TSC.\n");

    start = grub_get_time_ms();
    tsc1 = rdtsc();

    while (grub_get_time_ms() - start < 1000);

    stop = grub_get_time_ms();
    tsc2 = rdtsc();

    tsc_per_msec = grub_divmod64(tsc2 - tsc1, stop - start, 0);
    dprintf("testsuite", "TSC per ms = %llu\n", tsc_per_msec);

    grub_memset(bin, 0, sizeof(bin));

    bin[0].max = grub_divmod64(tsc_per_msec, 1000, 0);
    bin[1].max = grub_divmod64(tsc_per_msec, 100, 0);
    bin[2].max = grub_divmod64(tsc_per_msec, 10, 0);
    bin[3].max = tsc_per_msec;
    for (i = 4; i < LATENCY_BINS; i++)
        bin[i].max = bin[i - 1].max * 10;
    bin[LATENCY_BINS - 1].max = ~0ULL;

    bin[0].desc = "0     < t <=   1us";
    bin[1].desc = "1us   < t <=  10us";
    bin[2].desc = "10us  < t <= 100us";
    bin[3].desc = "100us < t <=   1ms";
    bin[4].desc = "1ms   < t <=  10ms";
    bin[5].desc = "10ms  < t <= 100ms";
    bin[6].desc = "100ms < t <=   1s ";
    bin[7].desc = "1s    < t <=  10s ";
    bin[8].desc = "10s   < t <= 100s ";
    bin[9].desc = "100s  < t         ";

    grub_printf("Starting pass #2.  Wait here, I will be back in 15 seconds.\n");
    test_latency(15 * 1000 * tsc_per_msec, &avg, &count, &max, bin, LATENCY_BINS);

    dprintf("testsuite", "%llu samples; average = %llu\n", count, avg);

    for (i = 0, bin_time = 1; i < LATENCY_BINS; i++, bin_time *= 10)
        if (bin[i].count) {
            grub_printf("%s; average = ", bin[i].desc);
            print_time(bin[i].avg, tsc_per_msec);
            grub_printf("; count = %llu\n Times between first few observations:", bin[i].count);
            for (j = 0; j < bin[i].recent_index - 1; j++) {
                grub_printf(" ");
                print_time(bin[i].recent_absolute[j + 1] - bin[i].recent_absolute[j], tsc_per_msec);
            }
            grub_printf("\n");
        }

    grub_printf("\nSummary of impact: observed maximum latency = ");
    print_time(max, tsc_per_msec);
    grub_printf("\n");
    max = grub_divmod64(max * 1000, tsc_per_msec, 0); /* Convert to microseconds */
    if (max < 150)
        grub_printf("Minimal risk of OS timeouts exists with this observed latency (< 150us).\n");
    else if (max > 300)
        grub_printf("Risk of OS timeouts exists with this observed latency (ideal < 150us).\n");
    else
        grub_printf("Potential risk of OS timeouts exists with this observed latency (ideal < 150us).\n");

    return GRUB_ERR_NONE;
}

static const struct grub_arg_option msr_available_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_QUIET
#define OPTION_QUIET 1
    {"quiet", 'q', 0, "No display (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_msr_available(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 cpuNum = viewpoint;
    U32 num;
    U32 i;
    bool status = true;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: MSR");
    if (strtou32_h(args[0], &num) != GRUB_ERR_NONE)
        return grub_errno;

    get_msr(cpuNum, num, ~0, 0);

    for (i = 0; i < unique_count; i++)
        status = status && !unique_msr[i].msr.status;

    if (!cmd->state[OPTION_QUIET].set) {
        if (status)
            grub_printf("Read MSR 0x%x is successful.\n", num);
        else
            grub_printf("Read MSR 0x%x causes GPF.\n", num);
    }

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const struct grub_arg_option pcie_read_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply to value read (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_VARNAME
#define OPTION_VARNAME 1
    {"varname", 'v', 0, "Save read value into variable VARNAME", "VARNAME", ARG_TYPE_STRING},
#undef OPTION_BYTES
#define OPTION_BYTES 2
    {"bytes", 'b', 0, "Bytes to read for PCI value", "BYTES", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 3
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
#undef OPTION_MEMADDR
#define OPTION_MEMADDR 5
    {"memaddr", 'p', 0, "PCIE memory base address (default=$pciexbase)", "MEMADDR", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_pcie_read(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 mask = ~0;
    U32 shift = 0;
    U32 bus, dev, fn, reg, size;
    U32 value = 0;
    U32 memaddr = 0;

    const char *varname;
    const char *pciexbase;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 4)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 4 arguments: bus, dev, fn, reg");
    if (strtou32_h(args[0], &bus) != GRUB_ERR_NONE ||
        strtou32_h(args[1], &dev) != GRUB_ERR_NONE ||
        strtou32_h(args[2], &fn) != GRUB_ERR_NONE ||
        strtou32_h(args[3], &reg) != GRUB_ERR_NONE)
        return grub_errno;

    size = (reg % 2) ? 1 :
           (reg % 4) ? 2 : 4;

    if (cmd->state[OPTION_BYTES].set)
        if (strtou32_h(cmd->state[OPTION_BYTES].arg, &size) != GRUB_ERR_NONE)
            return grub_errno;

    pciexbase = grub_env_get("pciexbase");
    if (pciexbase)
        if (strtou32_h(pciexbase, &memaddr) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MEMADDR].set)
        if (strtou32_h(cmd->state[OPTION_MEMADDR].arg, &memaddr) != GRUB_ERR_NONE)
            return grub_errno;

    if (!memaddr)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCIE memory base address");

    if (cmd->state[OPTION_VARNAME].set)
        varname = cmd->state[OPTION_VARNAME].arg;
    else
        varname = NULL;

    switch (size) {
    case 1:
        value = pci_mmcfg_read_u8(memaddr, bus, dev, fn, reg);
        break;
    case 2:
        value = pci_mmcfg_read_u16(memaddr, bus, dev, fn, reg);
        break;
    case 4:
        value = pci_mmcfg_read_u32(memaddr, bus, dev, fn, reg);
        break;
    default:
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCI read size (expected 1, 2, or 4)");
    }
    value = (value >> shift) & mask;

    if (!cmd->state[OPTION_QUIET].set) {
        if (mask == ~0U) {
            if (shift == 0)
                grub_printf("PCI %02x:%02x.%01x reg 0x%02x = 0x%x\n", bus, dev, fn, reg, value);
            else
                grub_printf("PCI %02x:%02x.%01x (reg 0x%02x) >> %u = 0x%x\n", bus, dev, fn, reg, shift, value);
        } else
            grub_printf("PCI %02x:%02x.%01x ((reg 0x%02x) >> %u) & 0x%x = 0x%x\n", bus, dev, fn, reg, shift, mask, value);
    }

    if (varname) {
        char buf[sizeof("0xXXXXXXXX")];
        grub_snprintf(buf, sizeof(buf), "0x%x", value);
        grub_env_set(varname, buf);
    }

    return GRUB_ERR_NONE;
}

static const struct grub_arg_option pcie_write_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply to values read and written (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_BYTES
#define OPTION_BYTES 1
    {"bytes", 'b', 0, "Bytes to read for PCI value", "BYTES", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 2
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_RMW
#define OPTION_RMW 3
    {"rmw", 'r', 0, "Read-modify-write operation", 0, 0},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
#undef OPTION_MEMADDR
#define OPTION_MEMADDR 5
    {"memaddr", 'p', 0, "PCIE memory base address (default=$pciexbase)", "MEMADDR", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_pcie_write(struct grub_extcmd *cmd, int argc, char **args)
{
    bool rmw;
    U32 mask = ~0;
    U32 shift = 0;
    U32 bus, dev, fn, reg, size;
    U32 value;
    U32 memaddr = 0;
    const char *pciexbase;

    rmw = cmd->state[OPTION_RMW].set;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 5)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 5 arguments: bus, dev, fn, reg, value");
    if (strtou32_h(args[0], &bus) != GRUB_ERR_NONE ||
        strtou32_h(args[1], &dev) != GRUB_ERR_NONE ||
        strtou32_h(args[2], &fn) != GRUB_ERR_NONE ||
        strtou32_h(args[3], &reg) != GRUB_ERR_NONE ||
        strtou32_h(args[4], &value) != GRUB_ERR_NONE)
        return grub_errno;

    size = (reg % 2) ? 1 :
           (reg % 4) ? 2 : 4;

    if (cmd->state[OPTION_BYTES].set)
        if (strtou32_h(cmd->state[OPTION_BYTES].arg, &size) != GRUB_ERR_NONE)
            return grub_errno;

    pciexbase = grub_env_get("pciexbase");
    if (pciexbase)
        if (strtou32_h(pciexbase, &memaddr) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MEMADDR].set)
        if (strtou32_h(cmd->state[OPTION_MEMADDR].arg, &memaddr) != GRUB_ERR_NONE)
            return grub_errno;

    if (!memaddr)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCIE memory base address");

    if (!cmd->state[OPTION_QUIET].set) {
        if (mask == ~0U) {
            if (shift == 0)
                grub_printf("PCI %02x:%02x.%01x reg 0x%02x %s 0x%x\n", bus, dev, fn, reg, rmw ? "|=" : "=", value);
            else
                grub_printf("PCI %02x:%02x.%01x reg 0x%02x %s 0x%x << %u  (0x%x)\n", bus, dev, fn, reg, rmw ? "|=" : "=", value, shift, value << shift);
        } else
            grub_printf("PCI %02x:%02x.%01x reg 0x%02x %s (0x%x & 0x%x) << %u  (0x%x)\n", bus, dev, fn, reg, rmw ? "|=" : "=", value, mask, shift, (value & mask) << shift);
    }

    value = (value & mask) << shift;
    switch (size) {
    case 1:
        if (rmw)
            value |= pci_mmcfg_read_u8(memaddr, bus, dev, fn, reg) & ~(mask << shift);
        pci_mmcfg_write_u8(memaddr, bus, dev, fn, reg, value);
        break;
    case 2:
        if (rmw)
            value |= pci_mmcfg_read_u16(memaddr, bus, dev, fn, reg) & ~(mask << shift);
        pci_mmcfg_write_u16(memaddr, bus, dev, fn, reg, value);
        break;
    case 4:
        if (rmw)
            value |= pci_mmcfg_read_u32(memaddr, bus, dev, fn, reg) & ~(mask << shift);
        pci_mmcfg_write_u32(memaddr, bus, dev, fn, reg, value);
        break;
    default:
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCI write size (expected 1, 2, or 4)");
    }

    return GRUB_ERR_NONE;
}

static const struct grub_arg_option pci_read_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply to value read (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_VARNAME
#define OPTION_VARNAME 1
    {"varname", 'v', 0, "Save read value into variable VARNAME", "VARNAME", ARG_TYPE_STRING},
#undef OPTION_BYTES
#define OPTION_BYTES 2
    {"bytes", 'b', 0, "Bytes to read for PCI value", "BYTES", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 3
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_pci_read(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 mask = ~0;
    U32 shift = 0;
    U32 bus, dev, fn, reg, size;
    U32 value = 0;

    const char *varname;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 4)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 4 arguments: bus, dev, fn, reg");
    if (strtou32_h(args[0], &bus) != GRUB_ERR_NONE ||
        strtou32_h(args[1], &dev) != GRUB_ERR_NONE ||
        strtou32_h(args[2], &fn) != GRUB_ERR_NONE ||
        strtou32_h(args[3], &reg) != GRUB_ERR_NONE)
        return grub_errno;

    size = (reg % 2) ? 1 :
           (reg % 4) ? 2 : 4;

    if (cmd->state[OPTION_BYTES].set)
        if (strtou32_h(cmd->state[OPTION_BYTES].arg, &size) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_VARNAME].set)
        varname = cmd->state[OPTION_VARNAME].arg;
    else
        varname = NULL;

    switch (size) {
    case 1:
        value = pci_read_u8(bus, dev, fn, reg);
        break;
    case 2:
        value = pci_read_u16(bus, dev, fn, reg);
        break;
    case 4:
        value = pci_read_u32(bus, dev, fn, reg);
        break;
    default:
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCI read size (expected 1, 2, or 4)");
    }
    value = (value >> shift) & mask;

    if (!cmd->state[OPTION_QUIET].set) {
        if (mask == ~0U) {
            if (shift == 0)
                grub_printf("PCI %02x:%02x.%01x reg 0x%02x = 0x%x\n", bus, dev, fn, reg, value);
            else
                grub_printf("PCI %02x:%02x.%01x (reg 0x%02x) >> %u = 0x%x\n", bus, dev, fn, reg, shift, value);
        } else
            grub_printf("PCI %02x:%02x.%01x ((reg 0x%02x) >> %u) & 0x%x = 0x%x\n", bus, dev, fn, reg, shift, mask, value);
    }

    if (varname) {
        char buf[sizeof("0xXXXXXXXX")];
        grub_snprintf(buf, sizeof(buf), "0x%x", value);
        grub_env_set(varname, buf);
    }

    return GRUB_ERR_NONE;
}

static const struct grub_arg_option pci_write_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply to values read and written (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_BYTES
#define OPTION_BYTES 1
    {"bytes", 'b', 0, "Bytes to read for PCI value", "BYTES", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 2
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_RMW
#define OPTION_RMW 3
    {"rmw", 'r', 0, "Read-modify-write operation", 0, 0},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_pci_write(struct grub_extcmd *cmd, int argc, char **args)
{
    bool rmw;
    U32 mask = ~0;
    U32 shift = 0;
    U32 bus, dev, fn, reg, size;
    U32 value;

    rmw = cmd->state[OPTION_RMW].set;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 5)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 5 arguments: bus, dev, fn, reg, value");
    if (strtou32_h(args[0], &bus) != GRUB_ERR_NONE ||
        strtou32_h(args[1], &dev) != GRUB_ERR_NONE ||
        strtou32_h(args[2], &fn) != GRUB_ERR_NONE ||
        strtou32_h(args[3], &reg) != GRUB_ERR_NONE ||
        strtou32_h(args[4], &value) != GRUB_ERR_NONE)
        return grub_errno;

    size = (reg % 2) ? 1 :
           (reg % 4) ? 2 : 4;

    if (cmd->state[OPTION_BYTES].set)
        if (strtou32_h(cmd->state[OPTION_BYTES].arg, &size) != GRUB_ERR_NONE)
            return grub_errno;

    if (!cmd->state[OPTION_QUIET].set) {
        if (mask == ~0U) {
            if (shift == 0)
                grub_printf("PCI %02x:%02x.%01x reg 0x%02x %s 0x%x\n", bus, dev, fn, reg, rmw ? "|=" : "=", value);
            else
                grub_printf("PCI %02x:%02x.%01x reg 0x%02x %s 0x%x << %u  (0x%x)\n", bus, dev, fn, reg, rmw ? "|=" : "=", value, shift, value << shift);
        } else
            grub_printf("PCI %02x:%02x.%01x reg 0x%02x %s (0x%x & 0x%x) << %u  (0x%x)\n", bus, dev, fn, reg, rmw ? "|=" : "=", value, mask, shift, (value & mask) << shift);
    }

    value = (value & mask) << shift;
    switch (size) {
    case 1:
        if (rmw)
            value |= pci_read_u8(bus, dev, fn, reg) & ~(mask << shift);
        pci_write_u8(bus, dev, fn, reg, value);
        break;
    case 2:
        if (rmw)
            value |= pci_read_u16(bus, dev, fn, reg) & ~(mask << shift);
        pci_write_u16(bus, dev, fn, reg, value);
        break;
    case 4:
        if (rmw)
            value |= pci_read_u32(bus, dev, fn, reg) & ~(mask << shift);
        pci_write_u32(bus, dev, fn, reg, value);
        break;
    default:
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCI write size (expected 1, 2, or 4)");
    }

    return GRUB_ERR_NONE;
}

static void noop_callback(void *param)
{
    (void)param;
}

static const struct grub_arg_option cpu_ping_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_cpu_ping(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 i, j;
    U32 cpuNum = viewpoint;
    U32 repeat_count;
    U64 start, stop;
    U32 seconds = 0;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: repeat_count");
    if (strtou32_h(args[0], &repeat_count) != GRUB_ERR_NONE)
        return grub_errno;

    start = grub_get_time_ms();
    for (j = 0; j < repeat_count; j++) {
        if (grub_checkkey() >= 0 && GRUB_TERM_ASCII_CHAR(grub_getkey()) == GRUB_TERM_ESC)
            break;
        stop = grub_get_time_ms();
        if (stop - start > 1000) {
            start = stop;
            seconds++;
            grub_printf("\r%u second%s (%u%%)", seconds, seconds == 1 ? "" : "s", (j * 100) / repeat_count);
        }
        for (i = first_cpu(cpuNum); i != last_cpu(cpuNum); i++)
            smp_function(cpu[i].apicid, noop_callback, NULL);
    }
    grub_printf("\r");

    return GRUB_ERR_NONE;
}

static void do_smp_sleep(void *param)
{
    U32 *usec = param;
    smp_sleep(*usec);
}

static grub_err_t grub_cmd_cpu_sleep(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 cpu_num;
    U32 usec;

    (void)cmd;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (argc != 2)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need a CPU and a time in usec");
    if (strtou32_h(args[0], &cpu_num) != GRUB_ERR_NONE)
        return grub_errno;
    if (strtou32_h(args[1], &usec) != GRUB_ERR_NONE)
        return grub_errno;

    smp_function(cpu[cpu_num].apicid, do_smp_sleep, &usec);

    return GRUB_ERR_NONE;
}

static const struct grub_arg_option format_options[] = {
#undef OPTION_DEC
#define OPTION_DEC 0
    {"dec", 'd', 0, "Format as decimal (default=enabled)", 0, 0},
#undef OPTION_HEX
#define OPTION_HEX 1
    {"hex", 'x', 0, "Format as hexadecimal", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_format(struct grub_extcmd *cmd, int argc, char **args)
{
    int i;
    bool status = true;

    if (!argc)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Requires format (decimal or hex) & list of environment variables");

    if (cmd->state[OPTION_HEX].set && cmd->state[OPTION_DEC].set)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "More than one format has been selected");

    for (i = 0; i < argc; i++) {
        char *str = grub_env_get(args[i]);
        U64 num;

        if (strtou64_h(str, &num) == GRUB_ERR_NONE) {
            char buf[sizeof("18446744073709551615")];
            if (cmd->state[OPTION_HEX].set)
                grub_snprintf(buf, sizeof(buf), "0x%llx", num);
            else
                grub_snprintf(buf, sizeof(buf), "%llu", num);
            grub_env_set(args[i], buf);
        } else
            status = false;
    }
    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static void mem_callback(void *param)
{
    struct mem *m = param;
    m->value = *(U32 *)m->addr;
}

static void get_mem(U32 cpu_num, U32 addr, U64 mask, U32 shift)
{
    U32 i, j;

    unique_count = 0;

    for (i = first_cpu(cpu_num); i != last_cpu(cpu_num); i++) {
        bool found = false;
        mem[i].addr = addr;
        if (!smp_function(cpu[i].apicid, mem_callback, &mem[i]))
            grub_printf("DEBUG: smp_function failed in get_mem\n");
        mem[i].value >>= shift;
        mem[i].value &= mask;

        for (j = 0; j < unique_count; j++)
            if (unique_mem[j].mem.value == mem[i].value) {
                unique_mem[j].count++;
                found = true;
                break;
            }
        if (!found) {
            unique_mem[unique_count].count = 1;
            unique_mem[unique_count].mem = mem[i];
            unique_count++;
        }
    }
}

static const struct grub_arg_option rdmem_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_VARNAME
#define OPTION_VARNAME 1
    {"varname", 'v', 0, "Save read value into variable VARNAME", "VARNAME", ARG_TYPE_STRING},
#undef OPTION_MASK
#define OPTION_MASK 2
    {"mask", 'm', 0, "Mask to apply to value (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 3
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_rdmem(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 cpuNum = viewpoint;
    U32 memAddr;
    const char *varname;
    U64 mask = ~0ULL;
    U32 shift = 0;
    U32 i, j;
    char mem_desc[128];
    bool status = true;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK].set)
        if (strtou64_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: memory_address");
    if (strtou32_h(args[0], &memAddr) != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_VARNAME].set)
        varname = cmd->state[OPTION_VARNAME].arg;
    else
        varname = NULL;

    get_mem(cpuNum, memAddr, mask, shift);

    // Set fail status for more than one unique value
    if (unique_count != 1)
        status = false;

    if (!cmd->state[OPTION_QUIET].set)
        for (i = 0; i < unique_count; i++) {
            if (mask == ~0ULL) {
                if (shift == 0)
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x", memAddr);
                else
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x shr %u", memAddr, shift);
            } else {
                if (shift == 0)
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x & 0x%llx", memAddr, mask);
                else
                    grub_snprintf(mem_desc, sizeof(mem_desc), "memory 0x%x shr %u & 0x%llx", memAddr, shift, mask);
            }

            grub_printf("%s = 0x%x\n", mem_desc, unique_mem[i].mem.value);

            grub_printf("On %u CPUs:%s", unique_mem[i].count, unique_mem[i].count == ncpus ? " all" : "");
            if (unique_mem[i].count != ncpus)
                for (j = first_cpu(cpuNum); j < last_cpu(cpuNum); j++)
                    if (unique_mem[i].mem.value == mem[j].value)
                        grub_printf(" 0x%x", cpu[j].apicid);
            grub_printf("\n");
        }

    if (varname) {
        char buf[sizeof("0xXXXXXXXXXXXXXXXX")];
        if (unique_count != 1) {
            if (!cmd->state[OPTION_QUIET].set)
                grub_printf("Variable setting requested but memory value not unique\n");
            return grub_error(GRUB_ERR_TEST_FAILURE, "false");
        }
        grub_snprintf(buf, sizeof(buf), "0x%x", unique_mem[0].mem.value);
        grub_env_set(varname, buf);
    }

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const struct grub_arg_option rdmsr_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_VARNAME
#define OPTION_VARNAME 1
    {"varname", 'v', 0, "Save read value into variable VARNAME", "VARNAME", ARG_TYPE_STRING},
#undef OPTION_MASK
#define OPTION_MASK 2
    {"mask", 'm', 0, "Mask to apply to value (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 3
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_rdmsr(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 cpuNum = viewpoint;
    U32 msrNum;
    const char *varname;
    U64 mask = ~0ULL;
    U32 shift = 0;
    U32 i, j;
    char msr_desc[128];
    bool status = true;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK].set)
        if (strtou64_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 argument: msr_number");
    if (strtou32_h(args[0], &msrNum) != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_VARNAME].set)
        varname = cmd->state[OPTION_VARNAME].arg;
    else
        varname = NULL;

    get_msr(cpuNum, msrNum, mask, shift);

    // Set fail for GPF on any MSR read
    for (i = 0; i < unique_count; i++)
        if (unique_msr[i].msr.status)
            status = false;

    // Set fail status for more than one unique value
    if (unique_count != 1)
        status = false;

    if (!cmd->state[OPTION_QUIET].set)
        for (i = 0; i < unique_count; i++) {
            if (mask == ~0ULL) {
                if (shift == 0)
                    grub_snprintf(msr_desc, sizeof(msr_desc), "MSR 0x%x", msrNum);
                else
                    grub_snprintf(msr_desc, sizeof(msr_desc), "MSR 0x%x shr %u", msrNum, shift);
            } else {
                if (shift == 0)
                    grub_snprintf(msr_desc, sizeof(msr_desc), "MSR 0x%x & 0x%llx", msrNum, mask);
                else
                    grub_snprintf(msr_desc, sizeof(msr_desc), "MSR 0x%x shr %u & 0x%llx", msrNum, shift, mask);
            }

            if (unique_msr[i].msr.status) {
                status = false;
                grub_printf("%s = GPF\n", msr_desc);
            } else
                grub_printf("%s = 0x%llx\n", msr_desc, unique_msr[i].msr.value);

            grub_printf("On %u CPUs:%s", unique_msr[i].count, unique_msr[i].count == ncpus ? " all" : "");
            if (unique_msr[i].count != ncpus)
                for (j = first_cpu(cpuNum); j < last_cpu(cpuNum); j++)
                    if (msr_match(unique_msr[i].msr, msr[j]))
                        grub_printf(" 0x%x", cpu[j].apicid);
            grub_printf("\n");
        }

    if (varname) {
        char buf[sizeof("0xXXXXXXXXXXXXXXXX")];
        if (unique_count != 1) {
            if (!cmd->state[OPTION_QUIET].set)
                grub_printf("Variable setting requested but MSR value not unique\n");
            return grub_error(GRUB_ERR_TEST_FAILURE, "false");
        }
        if (unique_msr[0].msr.status) {
            if (!cmd->state[OPTION_QUIET].set)
                grub_printf("Variable setting requested but MSR read caused GPF\n");
            return grub_error(GRUB_ERR_TEST_FAILURE, "false");
        }
        grub_snprintf(buf, sizeof(buf), "0x%llx", unique_msr[0].msr.value);
        grub_env_set(varname, buf);
    }

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const struct grub_arg_option set_mwait_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_NO_INT_BREAK_EVENT
#define OPTION_NO_INT_BREAK_EVENT 0
    {"no-int-break-event", 'i', 0, "Interrupt Break Event Disable (default=enabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_set_mwait(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 i;
    U32 cpuNum = viewpoint;
    bool use_mwait;
    U32 mwait_hint;
    U32 int_break_event = 1;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_NO_INT_BREAK_EVENT].set)
        int_break_event = 0;

    if (argc < 1)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need \"disable\" or \"enable\" as first argument");
    if (grub_strcmp(args[0], "disable") == 0) {
        if (argc != 1)
            return grub_error(GRUB_ERR_BAD_ARGUMENT, "\"disable\" takes no arguments");
        use_mwait = false;
        mwait_hint = 0;
    } else if (grub_strcmp(args[0], "enable") == 0) {
        if (argc != 2)
            return grub_error(GRUB_ERR_BAD_ARGUMENT, "\"enable\" requires an MWAIT hint value");
        use_mwait = true;
        if (strtou32_h(args[1], &mwait_hint) != GRUB_ERR_NONE)
            return grub_errno;
    } else
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need \"disable\" or \"enable\" as first argument");

    for (i = first_cpu(cpuNum); i != last_cpu(cpuNum); i++) {
        smp_set_mwait(cpu[i].apicid, use_mwait, mwait_hint, int_break_event);
        smp_function(cpu[i].apicid, noop_callback, NULL);
    }

    return GRUB_ERR_NONE;
}

static const struct grub_arg_option test_msr_consistency_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply before comparing (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 1
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_test_msr_consistency(struct grub_extcmd *cmd, int argc, char **args)
{
    U64 mask = ~0ULL;
    U32 shift = 0;
    char *teststr;
    U32 first;
    U32 last;
    U64 num; /* U64 so it cannot overflow if last == UINT32_MAX */
    bool status = true;

    if (cmd->state[OPTION_MASK].set)
        if (strtou64_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc < 2 || argc > 3)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 2 or 3 arguments: description, first MSR, optional last MSR (default=first)");
    teststr = args[0];
    if (strtou32_h(args[1], &first) != GRUB_ERR_NONE)
        return grub_errno;
    if (argc == 2)
        last = first;
    else if (strtou32_h(args[2], &last) != GRUB_ERR_NONE)
        return grub_errno;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    for (num = first; num <= last; num++) {
        status = status && msr_consistent(teststr, num, mask, shift);
        if (grub_checkkey() >= 0 && GRUB_TERM_ASCII_CHAR(grub_getkey()) == GRUB_TERM_ESC) {
            grub_printf("Aborting test.\n");
            break;
        }
    }

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const struct grub_arg_option test_msr_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply before comparing (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 1
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_test_msr(struct grub_extcmd *cmd, int argc, char **args)
{
    U64 mask = ~0ULL;
    U32 num;
    U32 shift = 0;
    U64 value;
    const char *teststr;
    bool status;

    if (cmd->state[OPTION_MASK].set)
        if (strtou64_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 3)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 3 arguments: description, MSR, value");
    teststr = args[0];
    if (strtou32_h(args[1], &num) != GRUB_ERR_NONE)
        return grub_errno;
    if (strtou64_h(args[2], &value) != GRUB_ERR_NONE)
        return grub_errno;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    status = msr_mask_value(teststr, num, mask, value, shift);

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const struct grub_arg_option test_msr_available_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_test_msr_available(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 cpuNum = viewpoint;
    U32 num;
    U32 i;
    const char *teststr;
    bool status = true;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 2)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 2 arguments: \"description\", MSR");
    teststr = args[0];
    if (strtou32_h(args[1], &num) != GRUB_ERR_NONE)
        return grub_errno;

    get_msr(cpuNum, num, ~0, 0);

    for (i = 0; i < unique_count; i++)
        status = status && !unique_msr[i].msr.status;

    if (status)
        passed++;
    else
        failed++;

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const struct grub_arg_option test_pci_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply before comparing (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 1
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_BYTES
#define OPTION_BYTES 2
    {"bytes", 'b', 0, "Bytes to read for PCI value", "BYTES", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_test_pci(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 mask = ~0;
    U32 shift = 0;
    U32 bus, dev, fn, reg, size, expected_value, regval = 0;
    const char *teststr;
    bool status;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 6)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 6 arguments: description, bus, dev, fn, reg, value");
    teststr = args[0];
    if (strtou32_h(args[1], &bus) != GRUB_ERR_NONE ||
        strtou32_h(args[2], &dev) != GRUB_ERR_NONE ||
        strtou32_h(args[3], &fn) != GRUB_ERR_NONE ||
        strtou32_h(args[4], &reg) != GRUB_ERR_NONE ||
        strtou32_h(args[5], &expected_value) != GRUB_ERR_NONE)
        return grub_errno;

    size = (reg % 2) ? 1 :
           (reg % 4) ? 2 : 4;

    if (cmd->state[OPTION_BYTES].set)
        if (strtou32_h(cmd->state[OPTION_BYTES].arg, &size) != GRUB_ERR_NONE)
            return grub_errno;

    if (size != 1 && size != 2 && size != 4)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCI read size (expected 1, 2, or 4)");

    status = pci_mask_value(teststr, bus, dev, fn, reg, size, mask, shift, expected_value, &regval);

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const struct grub_arg_option test_pcie_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply before comparing (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 1
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_BYTES
#define OPTION_BYTES 2
    {"bytes", 'b', 0, "Bytes to read for PCI value", "BYTES", ARG_TYPE_STRING},
#undef OPTION_MEMADDR
#define OPTION_MEMADDR 3
    {"memaddr", 'p', 0, "PCIE memory base address (default=$pciexbase)", "MEMADDR", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_test_pcie(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 mask = ~0;
    U32 shift = 0;
    U32 bus, dev, fn, reg, size, expected_value, regval = 0;
    const char *teststr;
    bool status;
    U32 memaddr = 0;
    const char *pciexbase;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 6)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 6 arguments: description, bus, dev, fn, reg, value");
    teststr = args[0];
    if (strtou32_h(args[1], &bus) != GRUB_ERR_NONE ||
        strtou32_h(args[2], &dev) != GRUB_ERR_NONE ||
        strtou32_h(args[3], &fn) != GRUB_ERR_NONE ||
        strtou32_h(args[4], &reg) != GRUB_ERR_NONE ||
        strtou32_h(args[5], &expected_value) != GRUB_ERR_NONE)
        return grub_errno;

    size = (reg % 2) ? 1 :
           (reg % 4) ? 2 : 4;

    if (cmd->state[OPTION_BYTES].set)
        if (strtou32_h(cmd->state[OPTION_BYTES].arg, &size) != GRUB_ERR_NONE)
            return grub_errno;

    if (size != 1 && size != 2 && size != 4)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCIe read size (expected 1, 2, or 4)");

    pciexbase = grub_env_get("pciexbase");
    if (pciexbase)
        if (strtou32_h(pciexbase, &memaddr) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MEMADDR].set)
        if (strtou32_h(cmd->state[OPTION_MEMADDR].arg, &memaddr) != GRUB_ERR_NONE)
            return grub_errno;

    if (!memaddr)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCIE memory base address");

    status = pcie_mask_value(teststr, memaddr, bus, dev, fn, reg, size, mask, shift, expected_value, &regval);

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static const struct grub_arg_option test_pci_consistency_options[] = {
#undef OPTION_MASK
#define OPTION_MASK 0
    {"mask", 'm', 0, "Mask to apply before comparing (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 1
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_BYTES
#define OPTION_BYTES 2
    {"bytes", 'b', 0, "Bytes to read for PCI value", "BYTES", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_test_pci_consistency(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 mask = ~0;
    U32 shift = 0;
    U32 bus, first, last, dev, fn, reg, size, expected_value, regval = 0;
    const char *teststr;
    bool status = true;

    if (cmd->state[OPTION_MASK].set)
        if (strtou32_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc < 6 || argc > 7)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 6 or 7 arguments: description, first bus, dev, fn, reg, value, [last bus]");
    teststr = args[0];
    if (strtou32_h(args[1], &first) != GRUB_ERR_NONE ||
        strtou32_h(args[2], &dev) != GRUB_ERR_NONE ||
        strtou32_h(args[3], &fn) != GRUB_ERR_NONE ||
        strtou32_h(args[4], &reg) != GRUB_ERR_NONE ||
        strtou32_h(args[5], &expected_value) != GRUB_ERR_NONE)
        return grub_errno;

    if (argc == 6)
        last = first;
    else if (strtou32_h(args[6], &last) != GRUB_ERR_NONE)
        return grub_errno;

    size = (reg % 2) ? 1 :
           (reg % 4) ? 2 : 4;

    if (cmd->state[OPTION_BYTES].set)
        if (strtou32_h(cmd->state[OPTION_BYTES].arg, &size) != GRUB_ERR_NONE)
            return grub_errno;

    if (size != 1 && size != 2 && size != 4)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Invalid PCI read size (expected 1, 2, or 4)");

    for (bus = first; bus <= last; bus++)
        status = status && pci_mask_value(teststr, bus, dev, fn, reg, size, mask, shift, expected_value, &regval);

    return status ? GRUB_ERR_NONE : grub_error(GRUB_ERR_TEST_FAILURE, "false");
}

static grub_err_t grub_cmd_testsuite(struct grub_extcmd *cmd, int argc, char **args)
{
    (void)cmd;
    (void)argc;
    (void)args;

    return GRUB_ERR_NONE;
}

static const char test_c_help[] =
    "Usage: test_c \"description\" \"C-style expression with space-separated tokens\"\n"
    "Evaluate a C expression and increment pass/fail count based on the result.\n"
    "Refer to \"c --help\" for details on C expression evaluation\n";

static grub_err_t grub_cmd_test_c(grub_command_t cmd, int argc, char **args)
{
    U64 result;
    (void)cmd;

    if (argc == 1 && grub_strcmp(args[0], "--help") == 0) {
        grub_printf("%s", test_c_help);
        return GRUB_ERR_NONE;
    }

    if (c_expr(argc - 1, &args[1], &result)) {
        if (!result) {
            failed++;
            if (verbose)
                grub_printf("%s %s\n", args[0], "FAIL");
            return grub_error(GRUB_ERR_TEST_FAILURE, "false");
        } else {
            passed++;
            if (verbose >= V_PASS)
                grub_printf("%s %s\n", args[0], "PASS");
            return GRUB_ERR_NONE;
        }
    } else
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Failed to parse C expression");
}

static const struct grub_arg_option test_options_options[] = {
#undef OPTION_VERBOSE
#define OPTION_VERBOSE 0
    {"verbose", 'v', 0, "Verbose output level (default=0 disabled)", "VERBOSE", ARG_TYPE_STRING},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_test_options(struct grub_extcmd *cmd, int argc, char **args)
{
    (void)argc;
    (void)args;
    if (cmd->state[OPTION_VERBOSE].set)
        if (strtou32_h(cmd->state[OPTION_VERBOSE].arg, &verbose) != GRUB_ERR_NONE)
            return grub_errno;

    return GRUB_ERR_NONE;
}

static grub_err_t grub_cmd_test_summary(struct grub_extcmd *cmd, int argc, char **args)
{
    (void)cmd;
    (void)argc;
    (void)args;
    grub_printf("Summary: %u passed, %u failed\n", passed, failed);
    passed = failed = 0;
    return GRUB_ERR_NONE;
}

static const struct grub_arg_option timer_options[] = {
#undef OPTION_VARNAME
#define OPTION_VARNAME 0
    {"varname", 'v', 0, "Save elapsed time in milliseconds into variable VARNAME", "VARNAME", ARG_TYPE_STRING},
#undef OPTION_QUIET
#define OPTION_QUIET 1
    {"quiet", 'q', 0, "No display of elapsed time (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_timer(struct grub_extcmd *cmd, int argc, char **args)
{
    U64 iterations;
    static U64 start = 0;
    U64 stop;
    U64 iterations_per_ms;
    const char *varname;

    if (argc < 1 || argc > 2)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 1 or 2 arguments: \"start\" or \"stop\" [iterations]");
    if (grub_strcmp(args[0], "start") == 0) {
        if (argc != 1)
            return grub_error(GRUB_ERR_BAD_ARGUMENT, "\"start\" takes no arguments");
        start = grub_get_time_ms();
    } else if (grub_strcmp(args[0], "stop") == 0) {
        if (start == 0)
            return grub_error(GRUB_ERR_BAD_ARGUMENT, "\"stop\" requires \"start\" on a prior call to start the timer");
        stop = grub_get_time_ms();

        if (!cmd->state[OPTION_QUIET].set) {
            if (argc == 2) {
                if (strtou64_h(args[1], &iterations) != GRUB_ERR_NONE)
                    return grub_errno;
                iterations_per_ms = grub_divmod64(iterations, stop - start, 0);
                grub_printf("elapsed time = %lld ms with %lld iteration/ms", stop - start, iterations_per_ms);
            } else
                grub_printf("elapsed time = %lld ms", stop - start);
        }

        if (cmd->state[OPTION_VARNAME].set) {
            // the 64-bit value for time (in milliseconds) will be stored as decimal
            char buf[sizeof("18446744073709551615")];
            varname = cmd->state[OPTION_VARNAME].arg;
            grub_snprintf(buf, sizeof(buf), "%llu", stop - start);
            grub_env_set(varname, buf);
        }
        start = 0;
    } else
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need \"start\" or \"stop\" as first argument");

    return GRUB_ERR_NONE;
}

struct rmw_msr {
    U32 num;
    U32 shift;
    U64 mask;
    U64 value;
    U32 rd_status;
    U32 wr_status;
    bool read_modify_write;
};

static void wrmsr_callback(void *param)
{
    struct rmw_msr *p = param;
    U64 rd_value, wr_value = 0;

    p->rd_status = 0;
    if (p->read_modify_write) {
        rdmsr64(p->num, &rd_value, &p->rd_status);
        if (p->rd_status)
            return;
        wr_value = rd_value & ~(p->mask << p->shift);
    }

    wr_value |= (p->value & p->mask) << p->shift;
    wrmsr64(p->num, wr_value, &p->wr_status);
}

static const struct grub_arg_option wrmsr_options[] = {
#undef OPTION_CPU
#define OPTION_CPU 0
    {"cpu", 'c', 0, "CPU number", "CPU", ARG_TYPE_STRING},
#undef OPTION_MASK
#define OPTION_MASK 1
    {"mask", 'm', 0, "Mask to apply to value (default=~0)", "MASK", ARG_TYPE_STRING},
#undef OPTION_SHIFT
#define OPTION_SHIFT 2
    {"shift", 's', 0, "Shift count for mask and value (default=0)", "SHIFT", ARG_TYPE_STRING},
#undef OPTION_RMW
#define OPTION_RMW 3
    {"rmw", 'r', 0, "Read-modify-write operation", 0, 0},
#undef OPTION_QUIET
#define OPTION_QUIET 4
    {"quiet", 'q', 0, "No display unless failures occur (default=disabled)", 0, 0},
    {0, 0, 0, 0, 0, 0}
};

static grub_err_t grub_cmd_wrmsr(struct grub_extcmd *cmd, int argc, char **args)
{
    U32 i;
    U32 num;
    U64 value;
    U64 mask = ~0ULL;
    U32 shift = 0;
    struct rmw_msr rmw;
    U32 cpuNum = viewpoint;
    U32 success = 0, rd_fail = 0, wr_fail = 0;

    if (init() != GRUB_ERR_NONE)
        return grub_errno;

    if (cmd->state[OPTION_CPU].set)
        if (parse_cpu_num(cmd->state[OPTION_CPU].arg, &cpuNum) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_MASK].set)
        if (strtou64_h(cmd->state[OPTION_MASK].arg, &mask) != GRUB_ERR_NONE)
            return grub_errno;

    if (cmd->state[OPTION_SHIFT].set)
        if (strtou32_h(cmd->state[OPTION_SHIFT].arg, &shift) != GRUB_ERR_NONE)
            return grub_errno;

    if (argc != 2)
        return grub_error(GRUB_ERR_BAD_ARGUMENT, "Need 2 arguments: msr_number value");
    if (strtou32_h(args[0], &num) != GRUB_ERR_NONE)
        return grub_errno;
    if (strtou64_h(args[1], &value) != GRUB_ERR_NONE)
        return grub_errno;

    rmw.num = num;
    rmw.value = value;
    rmw.mask = mask;
    rmw.shift = shift;
    rmw.read_modify_write = cmd->state[OPTION_RMW].set;

    for (i = first_cpu(cpuNum); i != last_cpu(cpuNum); i++) {
        if (!smp_function(cpu[i].apicid, wrmsr_callback, &rmw))
            grub_printf("DEBUG: smp_function failed in wrmsr\n");

        if (rmw.rd_status) {
            rd_fail++;
            cpu_status[i] = 1;
        } else if (rmw.wr_status) {
            wr_fail++;
            cpu_status[i] = 2;
        } else {
            success++;
            cpu_status[i] = 0;
        }
    }

    if (rd_fail || wr_fail || !cmd->state[OPTION_QUIET].set) {
        grub_printf("MSR 0x%x %s (0x%llx & 0x%llx) << %u  (0x%llx)\n", num, rmw.read_modify_write ? "|=" : "=", value, mask, shift, (value & mask) << shift);
        if (rd_fail) {
            grub_printf("Read MSR fail (GPF) on %u CPUs:%s", rd_fail, rd_fail == ncpus ? " all" : "");
            if (rd_fail != ncpus)
                for (i = first_cpu(cpuNum); i < last_cpu(cpuNum); i++)
                    if (cpu_status[i] == 1)
                        grub_printf(" 0x%x", cpu[i].apicid);
            grub_printf("\n");
        }
        if (wr_fail) {
            grub_printf("Write MSR fail (GPF) on %u CPUs:%s", wr_fail, wr_fail == ncpus ? " all" : "");
            if (wr_fail != ncpus)
                for (i = first_cpu(cpuNum); i < last_cpu(cpuNum); i++)
                    if (cpu_status[i] == 2)
                        grub_printf(" 0x%x", cpu[i].apicid);
            grub_printf("\n");
        }
        if (success) {
            grub_printf("Write MSR pass on %u CPUs:%s", success, success == ncpus ? " all" : "");
            if (success != ncpus)
                for (i = first_cpu(cpuNum); i < last_cpu(cpuNum); i++)
                    if (cpu_status[i] == 0)
                        grub_printf(" 0x%x", cpu[i].apicid);
            grub_printf("\n");
        }
    }

    return GRUB_ERR_NONE;
}

static grub_extcmd_t cmd_brandstring;
static grub_command_t cmd_c;
static grub_extcmd_t cmd_cpu, cmd_cpuid32;
static grub_extcmd_t cmd_cpu_freq;
static grub_extcmd_t cmd_cpu_latency;
static grub_extcmd_t cmd_cpu_ping;
static grub_extcmd_t cmd_cpu_sleep;
static grub_extcmd_t cmd_format;
static grub_extcmd_t cmd_msr_available;
static grub_extcmd_t cmd_pcie_read, cmd_pcie_write;
static grub_extcmd_t cmd_pci_read, cmd_pci_write;
static grub_extcmd_t cmd_rdmem;
static grub_extcmd_t cmd_rdmsr;
static grub_extcmd_t cmd_set_mwait;
static grub_extcmd_t cmd_testsuite, cmd_test_msr, cmd_test_msr_consistency, cmd_test_pci, cmd_test_pci_consistency, cmd_test_options, cmd_test_summary;
static grub_extcmd_t cmd_test_msr_available;
static grub_extcmd_t cmd_test_pcie;
static grub_extcmd_t cmd_test_cpuid_consistent;
static grub_command_t cmd_test_c;
static grub_extcmd_t cmd_timer;
static grub_extcmd_t cmd_wrmsr;

GRUB_MOD_INIT(testsuite)
{
  cmd_brandstring = grub_register_extcmd("brandstring", grub_cmd_brandstring, GRUB_COMMAND_FLAG_BOTH,
                                         "[-c cpu_num]",
                                         "Display brand string obtained via CPUID instruction",
                                         brandstring_options);
  cmd_c = grub_register_command("c", grub_cmd_c, "\"C-style expression with space-separated tokens\"", "Evaluate a C expression.");
  cmd_cpu = grub_register_extcmd("cpu", grub_cmd_cpu, GRUB_COMMAND_FLAG_BOTH,
                                 "[-l] cpu_num",
                                 "Set CPU number or optionally list CPU numbers",
                                 cpu_options);
  cmd_cpuid32 = grub_register_extcmd("cpuid32", grub_cmd_cpuid32, GRUB_COMMAND_FLAG_BOTH,
                                     "[-q] [-e] [-c cpu_num] [-m mask] [-s shift] function_num [index_num]",
                                     "Display registers returned by CPUID instructions",
                                     cpuid32_options);
  cmd_cpu_freq = grub_register_extcmd("cpu_freq", grub_cmd_cpu_freq, GRUB_COMMAND_FLAG_BOTH,
                                      0,
                                      "Compute processor frequency",
                                      0);
  cmd_cpu_latency = grub_register_extcmd("cpu_latency", grub_cmd_cpu_latency, GRUB_COMMAND_FLAG_BOTH,
                                         0,
                                         "Test cpu latency",
                                         0);
  cmd_format = grub_register_extcmd("format", grub_cmd_format, GRUB_COMMAND_FLAG_BOTH,
                                    "[-d | -x] variable [variable...]",
                                    "Format environment variable",
                                    format_options);
  cmd_msr_available = grub_register_extcmd("msr_available", grub_cmd_msr_available, GRUB_COMMAND_FLAG_BOTH,
                                           "[-c cpu_num] msr",
                                           "MSR available (does not GPF on read)",
                                           msr_available_options);
  cmd_cpu_ping = grub_register_extcmd("cpu_ping", grub_cmd_cpu_ping, GRUB_COMMAND_FLAG_BOTH,
                                      "[-c cpu_num] count",
                                      "Ping CPU",
                                      cpu_ping_options);
  cmd_cpu_sleep = grub_register_extcmd("cpu_sleep", grub_cmd_cpu_sleep, GRUB_COMMAND_FLAG_BOTH,
                                       "cpu_num microseconds",
                                       "Sleep on a specified CPU",
                                       0);
  cmd_pcie_read = grub_register_extcmd("pcie_read", grub_cmd_pcie_read, GRUB_COMMAND_FLAG_BOTH,
                                       "[-q] [-m mask] [-v VARNAME] [-s shift] [-b bytes] [-p memaddr] bus dev fn reg",
                                       "Read PCIE register",
                                        pcie_read_options);
  cmd_pcie_write = grub_register_extcmd("pcie_write", grub_cmd_pcie_write, GRUB_COMMAND_FLAG_BOTH,
                                        "[-q] [-r] [-m mask] [-s shift] [-b bytes] [-p memaddr] bus dev fn reg value",
                                        "Write PCIE register",
                                        pcie_write_options);
  cmd_pci_read = grub_register_extcmd("pci_read", grub_cmd_pci_read, GRUB_COMMAND_FLAG_BOTH,
                                      "[-q] [-m mask] [-v VARNAME] [-s shift] [-b bytes] bus dev fn reg",
                                      "Read PCI register",
                                      pci_read_options);
  cmd_pci_write = grub_register_extcmd("pci_write", grub_cmd_pci_write, GRUB_COMMAND_FLAG_BOTH,
                                      "[-q] [-r] [-m mask] [-s shift] [-b bytes] bus dev fn reg value",
                                      "Write PCI register",
                                      pci_write_options);
  cmd_rdmem = grub_register_extcmd("rdmem", grub_cmd_rdmem, GRUB_COMMAND_FLAG_BOTH,
                                   "[-q] [-c cpu_num] [-m mask] [-v VARNAME] [-s shift] memory_addr",
                                   "Read memory",
                                   rdmem_options);
  cmd_rdmsr = grub_register_extcmd("rdmsr", grub_cmd_rdmsr, GRUB_COMMAND_FLAG_BOTH,
                                   "[-q] [-c cpu_num] [-m mask] [-v VARNAME] [-s shift] msr_num",
                                   "Read MSR",
                                   rdmsr_options);
  cmd_set_mwait = grub_register_extcmd("set_mwait", grub_cmd_set_mwait, GRUB_COMMAND_FLAG_BOTH,
                                       "[-c cpu_num] [disable | [-i] enable hint_value]",
                                       "Set MWAIT disable/enable, hint, and interrupt break event",
                                       set_mwait_options);
  cmd_testsuite = grub_register_extcmd("testsuite", grub_cmd_testsuite, GRUB_COMMAND_FLAG_BOTH,
                                       0,
                                       "Experimental undocumented command",
                                       0);
  cmd_test_c = grub_register_command("test_c", grub_cmd_test_c, "\"description\" \"C-style expression with space-separated tokens\"",
                                     "Evaluate a C expression and increment pass/fail count based on the result.");
  cmd_test_cpuid_consistent = grub_register_extcmd("test_cpuid_consistent", grub_cmd_test_cpuid_consistent, GRUB_COMMAND_FLAG_BOTH,
                                                   "[-q] [-e] [-c cpu_num] [-m mask] [-s shift] function_num [index_num]",
                                                   "Test for consistent registers returned by CPUID instructions",
                                                   test_cpuid_consistent_options);
  cmd_test_msr = grub_register_extcmd("test_msr", grub_cmd_test_msr, GRUB_COMMAND_FLAG_BOTH,
                                      "[-m mask] [-s shift] description msr value",
                                      "Test MSR values",
                                      test_msr_options);
  cmd_test_msr_available = grub_register_extcmd("test_msr_available", grub_cmd_test_msr_available, GRUB_COMMAND_FLAG_BOTH,
                                                "[-c cpu_num] description msr",
                                                "Test if MSR available (does not GPF on read)",
                                                test_msr_available_options);
  cmd_test_msr_consistency = grub_register_extcmd("test_msr_consistency", grub_cmd_test_msr_consistency, GRUB_COMMAND_FLAG_BOTH,
                                                  "[-m mask] [-s shift] description first_msr [last_msr]",
                                                  "Test MSRs for consistency across CPUs",
                                                  test_msr_consistency_options);
  cmd_test_pci = grub_register_extcmd("test_pci", grub_cmd_test_pci, GRUB_COMMAND_FLAG_BOTH,
                                      "[-m mask] [-s shift] [-b bytes] description bus dev fn reg value",
                                      "Test PCI config space values",
                                      test_pci_options);
  cmd_test_pcie = grub_register_extcmd("test_pcie", grub_cmd_test_pcie, GRUB_COMMAND_FLAG_BOTH,
                                       "[-m mask] [-s shift] [-b bytes] [-p memaddr] description bus dev fn reg value",
                                       "Test PCIe config space values",
                                       test_pcie_options);
  cmd_test_pci_consistency = grub_register_extcmd("test_pci_consistency", grub_cmd_test_pci_consistency, GRUB_COMMAND_FLAG_BOTH,
                                                  "[-m mask] [-s shift] [-b bytes] description bus dev fn reg value [last_bus]",
                                                  "Test PCI config space values for consistency across buses",
                                                  test_pci_consistency_options);
  cmd_test_options = grub_register_extcmd("test_options", grub_cmd_test_options, GRUB_COMMAND_FLAG_BOTH,
                                          "[-v]",
                                          "Set test suite options",
                                          test_options_options);
  cmd_test_summary = grub_register_extcmd("test_summary", grub_cmd_test_summary, GRUB_COMMAND_FLAG_BOTH,
                                          0,
                                          "Summarize and reset test suite results",
                                          0);
  cmd_wrmsr = grub_register_extcmd("wrmsr", grub_cmd_wrmsr, GRUB_COMMAND_FLAG_BOTH,
                                   "[-q] [-c cpu_num] [-r] [-m mask] [-s shift] msr_num value",
                                   "Write MSR",
                                   wrmsr_options);
  cmd_timer = grub_register_extcmd("timer", grub_cmd_timer, GRUB_COMMAND_FLAG_BOTH,
                                   "start | [-v VARNAME] stop [count]",
                                   "Use test timer",
                                   timer_options);
}

GRUB_MOD_FINI(testsuite)
{
    grub_unregister_extcmd(cmd_wrmsr);
    grub_unregister_extcmd(cmd_timer);
    grub_unregister_extcmd(cmd_test_summary);
    grub_unregister_extcmd(cmd_test_options);
    grub_unregister_extcmd(cmd_test_pci_consistency);
    grub_unregister_extcmd(cmd_test_pcie);
    grub_unregister_extcmd(cmd_test_pci);
    grub_unregister_extcmd(cmd_test_msr_consistency);
    grub_unregister_extcmd(cmd_test_msr_available);
    grub_unregister_extcmd(cmd_test_msr);
    grub_unregister_extcmd(cmd_test_cpuid_consistent);
    grub_unregister_command(cmd_test_c);
    grub_unregister_extcmd(cmd_testsuite);
    grub_unregister_extcmd(cmd_set_mwait);
    grub_unregister_extcmd(cmd_rdmsr);
    grub_unregister_extcmd(cmd_rdmem);
    grub_unregister_extcmd(cmd_pci_write);
    grub_unregister_extcmd(cmd_pci_read);
    grub_unregister_extcmd(cmd_pcie_write);
    grub_unregister_extcmd(cmd_pcie_read);
    grub_unregister_extcmd(cmd_msr_available);
    grub_unregister_extcmd(cmd_format);
    grub_unregister_extcmd(cmd_cpu_sleep);
    grub_unregister_extcmd(cmd_cpu_ping);
    grub_unregister_extcmd(cmd_cpu_latency);
    grub_unregister_extcmd(cmd_cpu_freq);
    grub_unregister_extcmd(cmd_cpuid32);
    grub_unregister_extcmd(cmd_cpu);
    grub_unregister_command(cmd_c);
    grub_unregister_extcmd(cmd_brandstring);

    uninit();
}
